@@ -1,5 +1,5 @@
 /* idna.c - implementation of high-level IDNA processing function
-   Copyright (C) 2011-2016 Simon Josefsson
+   Copyright (C) 2011-2017 Simon Josefsson
 
    Libidn2 is free software: you can redistribute it and/or modify it
    under the terms of either:
@@ -36,13 +36,55 @@
 #include "bidi.h"
 #include "tables.h"
 #include "context.h"
+#include "tr46map.h"
 
+#include <unitypes.h>
 #include <unictype.h>		/* uc_is_general_category, UC_CATEGORY_M */
 #include <uninorm.h>		/* u32_normalize */
 #include <unistr.h>		/* u8_to_u32 */
 
 #include "idna.h"
 
+/*
+ * NFC Quick Check from
+ * http://unicode.org/reports/tr15/#Detecting_Normalization_Forms
+ *
+ * They say, this is much faster than 'brute force' normalization.
+ * Strings a very likely already in NFC form.
+ */
+static int
+_isNFC (uint32_t * label, size_t len)
+{
+  int lastCanonicalClass = 0;
+  int result = 1;
+  size_t it;
+
+  for (it = 0; it < len; it++)
+    {
+      uint32_t ch = label[it];
+
+      // supplementary code point
+      if (ch >= 0x10000)
+	it++;
+
+      int canonicalClass = uc_combining_class (ch);
+      if (lastCanonicalClass > canonicalClass && canonicalClass != 0)
+	return 0;
+
+      NFCQCMap *map = get_nfcqc_map (ch);
+      if (map)
+	{
+	  if (map->check)
+	    return 0;
+	  result = -1;
+	}
+
+      lastCanonicalClass = canonicalClass;
+    }
+
+  return result;
+}
+
 int
 _idn2_u8_to_u32_nfc (const uint8_t * src, size_t srclen,
 		     uint32_t ** out, size_t * outlen, int nfc)
@@ -58,7 +100,7 @@ _idn2_u8_to_u32_nfc (const uint8_t * src, size_t srclen,
       return IDN2_ENCODING_ERROR;
     }
 
-  if (nfc)
+  if (nfc && !_isNFC (p, plen))
     {
       size_t tmplen;
       uint32_t *tmp = u32_normalize (UNINORM_NFC, p, plen, NULL, &tmplen);
@@ -205,5 +247,34 @@ _idn2_label_test (int what, const uint32_t * label, size_t llen)
 	return rc;
     }
 
+  if (what & (TEST_TRANSITIONAL | TEST_NONTRANSITIONAL))
+    {
+      size_t i;
+      int transitional = what & TEST_TRANSITIONAL;
+
+      /* TR46: 4. The label must not contain a U+002E ( . ) FULL STOP */
+      for (i = 0; i < llen; i++)
+	if (label[i] == 0x002E)
+	  return IDN2_DOT_IN_LABEL;
+
+      /* TR46: 6. Each code point in the label must only have certain status
+       * values according to Section 5, IDNA Mapping Table:
+       *    a. For Transitional Processing, each value must be valid.
+       *    b. For Nontransitional Processing, each value must be either valid or deviation. */
+      for (i = 0; i < llen; i++)
+	{
+	  IDNAMap map;
+
+	  get_idna_map (label[i], &map);
+
+	  if (map_is (&map, TR46_FLG_VALID) ||
+	      (!transitional && map_is (&map, TR46_FLG_DEVIATION)))
+	    continue;
+
+	  return transitional ? IDN2_INVALID_TRANSITIONAL :
+	    IDN2_INVALID_NONTRANSITIONAL;
+	}
+    }
+
   return IDN2_OK;
 }
